@@ -1,16 +1,20 @@
 package Dancer2::Plugin::Auth::Extensible;
 
-our $VERSION = '0.622';
+our $VERSION = '0.700';
 
 use strict;
 use warnings;
 use Carp;
 use Dancer2::Core::Types qw(ArrayRef Bool HashRef Int Str);
+use Dancer2::FileUtils qw(path);
+use Dancer2::Template::Tiny;
+use File::Share qw(dist_dir);
 use List::Util qw(first);
 use Module::Runtime qw(use_module);
 use Scalar::Util;
 use Session::Token;
 use Try::Tiny;
+use URI::Escape;
 use Dancer2::Plugin;
 
 #
@@ -41,12 +45,24 @@ has login_page => (
     from_config => sub { '/login' },
 );
 
+has login_template => (
+    is          => 'ro',
+    isa         => Str,
+    from_config => sub { 'login' },
+);
+
 has login_page_handler => (
     is          => 'ro',
     isa         => Str,
     from_config => sub { '_default_login_page' },
 );
 
+has login_without_redirect => (
+    is          => 'ro',
+    isa         => Bool,
+    from_config => sub { 0 },
+);
+
 has logout_page => (
     is          => 'ro',
     isa         => Str,
@@ -185,6 +201,11 @@ has realm_providers => (
     init_arg => undef,
 );
 
+has _template_tiny => (
+    is      => 'ro',
+    default => sub { Dancer2::Template::Tiny->new },
+);
+
 #
 # hooks
 #
@@ -215,6 +236,8 @@ sub BUILD {
     my $plugin = shift;
     my $app    = $plugin->app;
 
+    Scalar::Util::weaken( my $weak_plugin = $plugin );
+
     warn "No Auth::Extensible realms configured with which to authenticate user"
       unless $plugin->realm_count;
 
@@ -230,8 +253,6 @@ sub BUILD {
         my $login_page  = $plugin->login_page;
         my $denied_page = $plugin->denied_page;
 
-        Scalar::Util::weaken( my $weak_plugin = $plugin );
-
         # Match optional reset code, but not "denied"
         $app->add_route(
             method => 'get',
@@ -254,10 +275,7 @@ sub BUILD {
                     && $weak_plugin->reset_password_handler
                     && $weak_plugin->user_password( code => $code ) )
                 {
-                    $app->request->params->{password_code_valid} = 1;
-                }
-                else {
-                    $app->response->status(401);
+                    $app->request->parameters->set('password_code_valid' => 1),
                 }
 
                 no strict 'refs';
@@ -297,6 +315,89 @@ sub BUILD {
             );
         }
     }
+
+    if ( $plugin->login_without_redirect ) {
+
+        # Add a post route so we can catch transparent login.
+        # This is a little sucky but since no hooks are called before
+        # route dispatch then adding this wildcard route now does at
+        # least make sure it gets added before any routes that use this
+        # plugin's route decorators are added.
+
+        $plugin->app->add_route(
+            method => 'post',
+            regexp => qr/.*/,
+            code   => sub {
+                my $app     = shift;
+                my $request = $app->request;
+
+                # See if this is actually a POST login.
+                my $username = $request->body_parameters->get(
+                    '__auth_extensible_username');
+
+                my $password = $request->body_parameters->get(
+                    '__auth_extensible_password');
+
+                if ( defined $username && defined $password ) {
+
+                    my $auth_realm = $request->body_parameters->get(
+                        '__auth_extensible_realm');
+
+                    # Remove the auth params since the forward we call later
+                    # will cause dispatch to retry this route again if
+                    # the original route was a post since dispatch starts
+                    # again from the start of the route list and this
+                    # wildcard route will get hit again causing a loop.
+                    foreach (qw/username password realm/) {
+                        $request->body_parameters->remove(
+                            "__auth_extensible_$_");
+                    }
+
+                    # Stash method and params since we delete these from
+                    # the session if login is successful but we still need
+                    # them for the forward to the original route after
+                    # success.
+                    my $method =
+                      $app->session->read('__auth_extensible_method');
+                    my $params =
+                      $app->session->read('__auth_extensible_params');
+
+                    # Attempt authentication.
+                    my ( $success, $realm ) =
+                      $weak_plugin->authenticate_user( $username,
+                        $password, $auth_realm );
+
+                    if ($success) {
+                        $app->session->delete('__auth_extensible_params');
+                        $app->session->delete('__auth_extensible_method');
+
+                        # Change session ID if we have a new enough D2
+                        # version with support.
+                        $app->change_session_id
+                          if $app->can('change_session_id');
+
+                        $app->session->write( logged_in_user => $username );
+                        $app->session->write( logged_in_user_realm => $realm );
+                        $app->log( core => "Realm is $realm" );
+                        $weak_plugin->execute_plugin_hook(
+                            'after_login_success');
+
+                    }
+                    else {
+                        $app->request->var( login_failed => 1 );
+                    }
+                    # Now forward to the original route using method and
+                    # params stashed in the session.
+                    $app->forward(
+                        $request->path,
+                        $params,
+                        { method => $method }
+                    );
+                }
+                $app->pass;
+            },
+        );
+    }
 }
 
 sub auth_provider {
@@ -508,6 +609,10 @@ sub logged_in_user {
 
 sub logged_in_user_lastlogin {
     my $lastlogin = shift->app->session->read('logged_in_user_lastlogin');
+    # We don't expect any bad $lastlogin values during testing so mark as many
+    # as possible as uncoverable.
+    # uncoverable branch false
+    # uncoverable condition right
     if ( defined $lastlogin && ref($lastlogin) eq '' && $lastlogin =~ /^\d+$/ )
     {
         # A sane epoch time. Old Provider::DBIC stores DateTime in the session
@@ -556,6 +661,8 @@ sub password_reset_send {
                 debug => "Failed to set_user_details with $realm: $_" );
         };
         if ($user) {
+            $plugin->app->log(
+                debug => "got one");
 
             # Okay, so email key is hard-coded, and therefore relies on the
             # provider returning that key. The alternative is to have a
@@ -586,6 +693,7 @@ sub require_any_role {
     return $plugin->_build_wrapper( @_, 'any' );
 }
 
+
 sub require_login {
     my $plugin  = shift;
     my $coderef = shift;
@@ -596,19 +704,11 @@ sub require_login {
                 warning => "Invalid require_login usage, please see docs" );
         }
 
-        my $user = $plugin->logged_in_user;
-        if ( !$user ) {
-            $plugin->execute_plugin_hook( 'login_required', $coderef );
+        # User already logged in so give them the page.
+        return $coderef->($plugin)
+          if $plugin->logged_in_user;
 
-            # TODO: see if any code executed by that hook set up a response
-            return $plugin->app->redirect(
-                $plugin->app->request->uri_for(
-                    $plugin->login_page,
-                    { return_url => $plugin->app->request->request_uri }
-                )
-            );
-        }
-        return $coderef->($plugin);
+        return $plugin->_check_for_login( $coderef );
     };
 }
 
@@ -699,6 +799,8 @@ sub user_password {
                     "Failed to check for code with $realm_check: $_" );
             };
             if ($username) {
+                $plugin->app->log( 'debug',
+                    "Found $username for code with $realm_check" );
                 $realm = $realm_check;
                 last;
             }
@@ -792,18 +894,8 @@ sub _build_wrapper {
       : $require_role;
 
     return sub {
-        my $user = $plugin->logged_in_user;
-        if ( !$user ) {
-            $plugin->execute_plugin_hook( 'login_required', $coderef );
-
-            # TODO: see if any code executed by that hook set up a response
-            return $plugin->app->redirect(
-                $plugin->app->request->uri_for(
-                    $plugin->login_page,
-                    { return_url => $plugin->app->request->request_uri }
-                )
-            );
-        }
+        return $plugin->_check_for_login( $coderef )
+          unless $plugin->logged_in_user;
 
         my $role_match;
 
@@ -840,15 +932,86 @@ sub _build_wrapper {
         $plugin->execute_plugin_hook( 'permission_denied', $coderef );
 
         # TODO: see if any code executed by that hook set up a response
-        return $plugin->app->redirect(
-            $plugin->app->request->uri_for(
-                $plugin->denied_page,
-                { return_url => $plugin->app->request->request_uri }
-            )
-        );
+
+        $plugin->app->response->status(403);
+        my $options;
+        my $view            = $plugin->denied_page;
+        my $template_engine = $plugin->app->template_engine;
+        my $path            = $template_engine->view_pathname($view);
+        if ( !-f $path ) {
+            $plugin->app->log(
+                debug => "app has no denied_page template defined" );
+            $options->{content} = $plugin->_render_template('login_denied.tt');
+            undef $view;
+        }
+        return $plugin->app->template( $view, undef, $options );
     };
 }
 
+sub _check_for_login {
+    my ( $plugin, $coderef ) = @_;
+    $plugin->execute_plugin_hook( 'login_required', $coderef );
+
+    # TODO: see if any code executed by that hook set up a response
+
+    my $request = $plugin->app->request;
+
+    if ( $plugin->login_without_redirect ) {
+        my $tokens = {
+            login_failed           => $request->var('login_failed'),
+            reset_password_handler => $plugin->reset_password_handler
+        };
+
+        # The WWW-Authenticate header added varies depending on whether
+        # the client is a robot or not.
+        my $ua = use_module('HTTP::BrowserDetect')
+          ->new( $request->env->{HTTP_USER_AGENT} );
+        my $base = $request->base;
+        my $auth_method;
+
+        if ( !$ua->browser_string || $ua->robot ) {
+            $auth_method = $auth_method = qq{Basic realm="$base"};
+        }
+        else {
+            $auth_method = qq{FormBasedLogin realm="$base", }
+              . q{comment="use form to log in"};
+        }
+
+        $plugin->app->response->status(401);
+        $plugin->app->response->push_header(
+            'WWW-Authenticate' => $auth_method );
+
+        # If this is the first attempt to reach a protected page and *not*
+        # a failed passthrough login then we need to stash method and params.
+        if ( !$request->var('login_failed') ) {
+            $plugin->app->session->write(
+                '__auth_extensible_method' => lc($request->method) );
+            $plugin->app->session->write(
+                '__auth_extensible_params' => \%{ $request->params } );
+        }
+
+        # If app has its own login page view then use it
+        # otherwise render our internal one and pass that to 'template'.
+        my ( $view, $options ) = ( $plugin->login_template, {} );
+        my $template_engine = $plugin->app->template_engine;
+        my $path            = $template_engine->view_pathname($view);
+        if ( !-f $path ) {
+            $plugin->app->log( debug => "app has no login template defined" );
+            $options->{content} =
+              $plugin->_render_template( 'transparent_login.tt', $tokens );
+            undef $view;
+        }
+        return $plugin->app->template( $view, $tokens, $options );
+    }
+
+    # old-fashioned redirect to login page with return_url set
+    return $plugin->app->redirect(
+        $request->uri_for(
+            $plugin->login_page, { return_url => $request->request_uri }
+        )
+    );
+}
+
 sub _default_email_password_reset {
     my ( $plugin, %options ) = @_;
 
@@ -873,91 +1036,56 @@ __EMAIL
     $plugin->_send_email( to => $options{email}, %message );
 }
 
+sub _render_template {
+    my ( $plugin, $view, $tokens ) = @_;
+    $tokens ||= +{};
+
+    my $template =
+      path( dist_dir('Dancer2-Plugin-Auth-Extensible'), 'views', $view );
+
+    $plugin->_template_tiny->render( $template, $tokens );
+}
+
 sub _default_login_page {
-    my $plugin    = shift;
-    my $app       = $plugin->app;
-    my $loginpage = $plugin->login_page;
+    my $plugin = shift;
+    my $request = $plugin->app->request;
 
-    if ( my $new_password = $app->request->param('new_password') ) {
-        return <<NEWPW;
-<h1>New password</h1>
-<p>
-Your new password is $new_password
-</p>
-<a href="$loginpage">Click here to login</a>
-NEWPW
+    # Simple escape of new_password param.
+    # This only works with the default password generator but since we
+    # are planning to remove password generation in favour of user-specified
+    # password on reset then this will do for now.
+    my $new_password = $request->parameters->get('new_password');
+    if ( defined $new_password ) {
+        $new_password =~ s/[^a-zA-Z0-9]//g;
     }
 
-    if ( $app->request->param('reset_sent') ) {
-        return <<SENT;
-<h1>Request sent</h1>
-<p>A password reset request has been sent. Please check your email.</p>
-SENT
+    # Make sure all tokens are escaped in some way.
+    my $tokens = {
+        loginpage    => uri_escape( $plugin->login_page ),
+        login_failed => !!$request->var('login_failed'),
+        new_password => $new_password,
+        password_code_valid =>
+          !!$request->parameters->get('password_code_valid'),
+        reset_sent             => !!$request->parameters->get('reset_sent'),
+        reset_password_handler => !!$plugin->reset_password_handler,
+        return_url => uri_escape( $request->parameters->get('return_url') ),
+    };
+
+    # If app has its own login page view then use it
+    # otherwise render our internal one and pass that to 'template'.
+    my ( $view, $options ) = ( $plugin->login_template, {} );
+    my $template_engine = $plugin->app->template_engine;
+    my $path            = $template_engine->view_pathname($view);
+    if ( !-f $path ) {
+        $plugin->app->log( debug => "app has no login template defined" );
+        $options->{content} = $plugin->_render_template( 'login.tt', $tokens );
+        undef $view;
     }
-
-    # Valid password reset request. Just need to confirm to
-    # prevent GET requests by email filters
-    if ( $app->request->param('password_code_valid') ) {
-        return <<VALID;
-<h1>Reset your password</h1>
-<p>
-Please click the button below to reset your password
-</p>
-<form method="post">
-<input type="submit" name="confirm_reset" value="Reset password">
-</form>
-VALID
-    }
-
-    my $pwreset_html =
-      !$plugin->reset_password_handler
-      ? ""
-      : <<RESETPW;
-<h2>Password reset</h2>
-<p>Enter your username to obtain an email to reset your password</p>
-<label for="username_reset">Username:</label>
-<input type="text" name="username_reset" id="username_reset">
-<input type="submit" name="submit_reset" value="Submit">
-RESETPW
-    my $return_url = $app->request->params->{return_url} || '';
-
-    my $login_fail_message =
-      $app->request->vars->{login_failed}
-      ? "<p>LOGIN FAILED</p>"
-      : "";
-
-    return <<PAGE;
-<h1>Login Required</h1>
-
-<p>
-You need to log in to continue.
-</p>
-
-$login_fail_message
-
-<form method="post">
-<label for="username">Username:</label>
-<input type="text" name="username" id="username">
-<br />
-<label for="password">Password:</label>
-<input type="password" name="password" id="password">
-<br />
-<input type="hidden" name="return_url" value="$return_url">
-<input type="submit" value="Login">
-$pwreset_html
-</form>
-PAGE
+    return $plugin->app->template( $view, $tokens, $options );
 }
 
 sub _default_permission_denied_page {
-    my $plugin = shift;
-    return <<PAGE
-<h1>Permission Denied</h1>
-
-<p>
-Sorry, you're not allowed to access that page.
-</p>
-PAGE
+    shift->_render_template( 'login_denied.tt' );
 }
 
 sub _default_welcome_send {
@@ -1051,8 +1179,8 @@ sub _logout_route {
 
     $app->destroy_session;
 
-    if ( $req->params->{return_url} ) {
-        $app->redirect( $req->params->{return_url} );
+    if ( my $url = $req->parameters->get('return_url') ) {
+        $app->redirect( $url );
     }
     elsif ($plugin->exit_page) {
         $app->redirect($plugin->exit_page);
@@ -1068,12 +1196,11 @@ sub _logout_route {
 sub _post_login_route {
     my $app = shift;
     my $plugin = $app->with_plugin('Auth::Extensible');
+    my $params = $app->request->body_parameters->as_hashref;
 
     # First check for password reset request, if applicable
-    if (   $plugin->reset_password_handler
-        && $app->request->param('submit_reset') )
-    {
-        my $username = $app->request->param('username_reset');
+    if (   $plugin->reset_password_handler && $params->{submit_reset} ) {
+        my $username = $params->{username_reset};
         croak "Attempt to pass reference to reset blocked" if ref $username;
         $plugin->password_reset_send( username => $username );
         $app->forward(
@@ -1086,7 +1213,7 @@ sub _post_login_route {
     # Then for a password reset itself (confirmed by POST request)
     my ($code) =
          $plugin->reset_password_handler
-      && $app->request->param('confirm_reset')
+      && $params->{confirm_reset}
       && $app->request->splat;
 
     if ($code) {
@@ -1111,8 +1238,8 @@ sub _post_login_route {
     # with paremeterisation) - but if params->{password} was something
     # different, e.g. { 'like' => '%' }, we might end up with some SQL like
     # WHERE password LIKE '%' instead - which would not be a Good Thing.
-    my ( $username, $password ) =
-      @{ $app->request->params() }{qw(username password)};
+    my $username = $params->{username} || $params->{__auth_extensible_username};
+    my $password = $params->{password} || $params->{__auth_extensible_password};
 
     for ( $username, $password ) {
         if ( ref $_ ) {
@@ -1124,11 +1251,11 @@ sub _post_login_route {
 
     if ( $plugin->logged_in_user ) {
         # uncoverable condition false
-        $app->redirect( $app->request->params->{return_url}
+        $app->redirect( $app->request->parameters->get('return_url')
               || $plugin->user_home_page );
     }
 
-    my $auth_realm = $app->request->param('realm');
+    my $auth_realm = $params->{realm} || $params->{__auth_extensible_realm};
     my ( $success, $realm ) =
       $plugin->authenticate_user( $username, $password, $auth_realm );
 
@@ -1143,7 +1270,7 @@ sub _post_login_route {
         $app->log( core => "Realm is $realm" );
         $plugin->execute_plugin_hook( 'after_login_success' );
         # uncoverable condition false
-        $app->redirect( $app->request->params->{return_url}
+        $app->redirect( $app->request->parameters->get('return_url')
               || $plugin->user_home_page );
     }
     else {
@@ -1291,9 +1418,7 @@ required methods, and you're good to go!
 
 Keywords are provided to check if a user is logged in / has appropriate roles.
 
-=over
-
-=item require_login - require the user to be logged in
+=head2 require_login - require the user to be logged in
 
     get '/dashboard' => require_login sub { .... };
 
@@ -1301,7 +1426,7 @@ If the user is not logged in, they will be redirected to the login page URL to
 log in.  The default URL is C</login> - this may be changed with the
 C<login_page> option.
 
-=item require_role - require the user to have a specified role
+=head2 require_role - require the user to have a specified role
 
     get '/beer' => require_role BeerDrinker => sub { ... };
 
@@ -1313,7 +1438,7 @@ to the access denied URL.
 If C<disable_roles> configuration option is set to a true value then using
 L</require_role> will cause the application to croak on load.
 
-=item require_any_roles - require the user to have one of a list of roles
+=head2 require_any_role - require the user to have one of a list of roles
 
     get '/drink' => require_any_role [qw(BeerDrinker VodaDrinker)] => sub {
         ...
@@ -1327,7 +1452,7 @@ roles, they will be redirected to the access denied URL.
 If C<disable_roles> configuration option is set to a true value then using
 L</require_any_roles> will cause the application to croak on load.
 
-=item require_all_roles - require the user to have all roles listed
+=head2 require_all_roles - require the user to have all roles listed
 
     get '/foo' => require_all_roles [qw(Foo Bar)] => sub { ... };
 
@@ -1339,9 +1464,104 @@ redirected to the access denied URL.
 If C<disable_roles> configuration option is set to a true value then using
 L</require_all_roles> will cause the application to croak on load.
 
-=back
+=head1 NO-REDIRECT LOGIN
 
-=head2 Replacing the Default C< /login > and C< /login/denied > Routes
+By default when a page is requested that requires login and the user is not
+logged in then the plugin redirects the user to the L</login_page> and sets
+C<return_url> to the page originally requested. After successful login the
+user is redirected to the originally-requested page.
+
+As an alternative if L</login_without_redirect> is true then the login
+process happens with no redirects. Instead a C<401> C<Unauthorized> code
+is returned and a login page is displayed. This login page is posted to the
+original URI and on successful login an internal L<Dancer2::Manual/forward>
+is performed so that the originally requested page is displayed. Any
+L<Dancer2::Manual/params> from the original request are added to the
+forward so that they are available to the page's route handler either using
+L<Dancer2::Manual/params> or L<Dancer2::Manual/query_parameters>.
+
+This relies on the login form having no C<action> set and also it must use
+C<__auth_extensible_username> and C<__auth_extensible_password> input names.
+Optionally  C<__auth_extensible_realm> can also be used in a custom login
+page.
+
+See L<http://shadow.cat/blog/matt-s-trout/humane-login-screens/> for the
+original idea for this functionality.
+
+=head1 CUSTOMISING C</login> AND C</login/denied>
+
+=head2 login_template
+
+The L</login_template> setting determines the name of the view you use
+for your custom login page. If this view exists in your application then it
+will be used instead of the default login template.
+
+If you are using L</login_without_redirect> and assuming you are using
+L<Template::Toolkit> then your custom login page should be something like this:
+
+    <h1>Login Required</h1>
+
+    <p>You need to log in to continue.</p>
+
+    [%- IF login_failed -%]
+        <p>LOGIN FAILED</p>
+    [%- END -%]
+
+    <form method="post">
+        <label for="username">Username:</label>
+        <input type="text" name="__auth_extensible_username" id="username">
+        <br />
+        <label for="password">Password:</label>
+        <input type="password" name="__auth_extensible_password" id="password">
+        <br />
+        <input type="submit" value="Login">
+    </form>
+
+    [%- IF reset_password_handler -%]
+    <form method="post" action="[% login_page %]">
+        <h2>Password reset</h2>
+        <p>Enter your username to obtain an email to reset your password</p>
+        <label for="username_reset">Username:</label>
+        <input type="text" name="username_reset" id="username_reset">
+        <input type="submit" name="submit_reset" value="Submit">
+    </form>
+    [%- END -%]
+
+If you are B<not> using L</login_without_redirect> and assuming you are using
+L<Template::Toolkit> then your custom login page should be something like this:
+
+    <h1>Login Required</h1>
+
+    <p>You need to log in to continue.</p>
+
+    [%- IF login_failed -%]
+        <p>LOGIN FAILED</p>
+    [%- END -%]
+
+    <form method="post">
+        <label for="username">Username:</label>
+        <input type="text" name="username" id="username">
+        <br />
+        <label for="password">Password:</label>
+        <input type="password" name="password" id="password">
+        <br />
+        <input type="submit" value="Login">
+
+        [%- IF return_url -%]
+            <input type="hidden" name="return_url" value="[% return_url %]">
+        [%- END -%]
+
+        [%- IF reset_password_handler -%]
+            <h2>Password reset</h2>
+            <p>Enter your username to obtain an email to reset your password</p>
+            <label for="username_reset">Username:</label>
+            <input type="text" name="username_reset" id="username_reset">
+            <input type="submit" name="submit_reset" value="Submit">
+        [%- END -%]
+
+    </form>
+
+=head2 Replacing the default C< /login > and C< /login/denied > routes
 
 By default, the plugin adds a route to present a simple login form at that URL.
 If you would rather add your own, set the C<no_default_pages> setting to a true
@@ -1403,51 +1623,19 @@ you can configure them. See below.
 The default routes also contain functionality for a user to perform password
 resets. See the L<PASSWORD RESETS> documentation for more details.
 
-=head2 Keywords
+=head1 KEYWORDS
 
-=over
+The following keywords are provided in additional to the route decorators
+specified in L</CONTROLLING ACCESS TO ROUTES>:
 
-=item require_login
+=head2 logged_in_user
 
-Used to wrap a route which requires a user to be logged in order to access
-it.
-
-    get '/secret' => require_login sub { .... };
-
-=item require_role
-
-Used to wrap a route which requires a user to be logged in as a user with the
-specified role in order to access it.
-
-    get '/beer' => require_role BeerDrinker => sub { ... };
-
-You can also provide a regular expression, if you need to match the role using a
-regex - for example:
-
-    get '/beer' => require_role qr/Drinker$/ => sub { ... };
-
-=item require_any_role
-
-Used to wrap a route which requires a user to be logged in as a user with any
-one (or more) of the specified roles in order to access it.
-
-    get '/foo' => require_any_role [qw(Foo Bar)] => sub { ... };
-
-=item require_all_roles
-
-Used to wrap a route which requires a user to be logged in as a user with all
-of the roles listed in order to access it.
-
-    get '/foo' => require_all_roles [qw(Foo Bar)] => sub { ... };
-
-
-=item logged_in_user
-
-Returns a hashref of details of the currently logged-in user, if there is one.
+Returns a hashref of details of the currently logged-in user or some kind of
+user object, if there is one.
 
 The details you get back will depend upon the authentication provider in use.
 
-=item get_user_details
+=head2 get_user_details
 
 Returns a hashref of details of the specified user. The realm can optionally
 be specified as the second parameter. If the realm is not specified, each
@@ -1455,7 +1643,7 @@ realm will be checked, and the first matching user will be returned.
 
 The details you get back will depend upon the authentication provider in use.
 
-=item user_has_role
+=head2 user_has_role
 
 Check if a user has the role named.
 
@@ -1471,7 +1659,7 @@ You can also provide the username to check;
 If C<disable_roles> configuration option is set to a true value then using
 L</user_has_role> will cause the application to croak at runtime.
 
-=item user_roles
+=head2 user_roles
 
 Returns a list of the roles of a user.
 
@@ -1483,7 +1671,7 @@ Returns a list or arrayref depending on context.
 If C<disable_roles> configuration option is set to a true value then using
 L</user_roles> will cause the application to croak at runtime.
 
-=item authenticate_user
+=head2 authenticate_user
 
 Usually you'll want to let the built-in login handling code deal with
 authenticating users, but in case you need to do it yourself, this keyword
@@ -1504,7 +1692,7 @@ you can supply the realm as an optional third parameter.
 In boolean context, returns simply true or false; in list context, returns
 C<($success, $realm)>.
 
-=item logged_in_user_lastlogin
+=head2 logged_in_user_lastlogin
 
 Returns (as a DateTime object) the time of the last successful login of the
 current logged in user.
@@ -1513,7 +1701,7 @@ To enable this functionality, set the configuration key C<record_lastlogin> to
 a true value. The backend provider must support write access for a user and
 have lastlogin functionality implemented.
 
-=item update_user
+=head2 update_user
 
 Updates a user's details. If the authentication provider supports it, this
 keyword allows a user's details to be updated within the backend data store.
@@ -1535,7 +1723,7 @@ Otherwise, the realm must be specified with the realm key.
 
 The updated user's details are returned, as per L<logged_in_user>.
 
-=item update_current_user
+=head2 update_current_user
 
 The same as L<update_user>, but does not take a username as the first parameter,
 instead updating the currently logged-in user.
@@ -1545,7 +1733,7 @@ instead updating the currently logged-in user.
 
 The updated user's details are returned, as per L<logged_in_user>.
 
-=item create_user
+=head2 create_user
 
 Creates a new user, if the authentication provider supports it. Optionally
 sends a welcome message with a password reset request, in which case an
@@ -1599,7 +1787,7 @@ L<password_reset_send_email>.
 
 =back
 
-=item password_reset_send
+=head2 password_reset_send
 
 L</password_reset_send> sends a user an email with a password reset link. Along
 with L</user_password>, it allows a user to reset their password.
@@ -1699,7 +1887,7 @@ Here is an example subroutine:
 
 =back
 
-=item user_password
+=head2 user_password
 
 This provides various functions to check or reset a user's password, either
 from a reset code that was previously send by L<password_reset_send> or
@@ -1743,7 +1931,7 @@ Force set a specific user's password, without checking existing password:
 
     user_password username => 'jbloggs', new_password => 'secret'
 
-=item logged_in_user_password_expired
+=head2 logged_in_user_password_expired
 
 Returns true if the password of the currently logged in user has expired.  To
 use this functionality, the provider must support the C<password_expired>
@@ -1765,8 +1953,6 @@ that a check is done in the C<before> hook:
         }
     }
 
-=back
-
 =head2 PASSWORD RESETS
 
 A variety of functionality is provided to make it easier to manage requests
@@ -1831,6 +2017,10 @@ In your application's configuation file:
             # will cause app to croak on load. Use of 'user_roles' and
             # 'user_has_role' will croak at runtime.
             disable_roles: 0
+            # Set to 1 to use the no-redirect login functionality
+            login_without_redirect: 0
+            # Set the view name for a custom login page, defaults to 'login'
+            login_template: login
             # After /login: If no return_url is given: land here ('/' is default)
             user_home_page: '/user'
             # After /logout: If no return_url is given: land here (no default)
@@ -1876,7 +2066,7 @@ the currently logged in user.
 Please see L<Dancer2::Core::Session> for information on how to configure session 
 management within your application.
 
-=head1 FUNCTIONS
+=head1 METHODS
 
 =head2 auth_provider($dsl, $realm)
 
@@ -1977,6 +2167,8 @@ Evan Brown (GH #20, #32).
 
 Jason Lewis (Unix provider problem).
 
+Matt S. Trout (mst) for L<Zero redirect login the easy and friendly way|http://shadow.cat/blog/matt-s-trout/humane-login-screens/>.
+
 =head1 LICENSE AND COPYRIGHT
 
 Copyright 2012-16 David Precious.
