@@ -1,6 +1,6 @@
 package Dancer2::Plugin::Auth::Extensible::Test;
 
-our $VERSION = '0.622';
+our $VERSION = '0.700';
 
 =head1 NAME
 
@@ -52,7 +52,6 @@ my %dispatch = (
     logged_in_user_lastlogin        => \&_logged_in_user_lastlogin,
     logged_in_user_password_expired => \&_logged_in_user_password_expired,
     password_reset                  => \&_password_reset,
-    password_reset_send             => \&_password_reset_send,
     require_login                   => \&_require_login,
     roles                           => \&_roles,
     update_current_user             => \&_update_current_user,
@@ -70,8 +69,7 @@ my %dependencies = (
     logged_in_user_lastlogin => ['create_user','record_lastlogin'],
     logged_in_user_password_expired =>
       [ 'get_user_details', 'password_expired' ],
-    password_reset      => ['get_user_by_code'],
-    password_reset_send => ['set_user_details'],
+    password_reset      => ['get_user_by_code', 'set_user_details'],
     require_login       => ['get_user_details'],
     roles               => ['get_user_roles' ],
     update_current_user => ['set_user_details'],
@@ -572,6 +570,8 @@ sub _create_user {
 
     # create user with `email_welcome` so we can test reset code
 
+    $Dancer2::Plugin::Auth::Extensible::Test::App::data = undef;
+
     $res = post(
         "/create_user",
         [
@@ -757,23 +757,31 @@ sub _login_logout {
     # get /login
 
     $res = get('/login');
-    TODO: {
-        local $TODO = "GET /login returns 401";
-        ok $res->is_success, "GET /login is_success";
-    }
+    ok $res->is_success, "GET /login is_success";
     like $res->content, qr/input.+name="password"/,
       "... and we have the login page.";
-    like $res->content,
-      qr/Enter your username to obtain an email to reset your password/,
-      "... which has password reset option (reset_password_handler=>1).";
+
+    # login page has password reset section (or not)
+    if ( grep { $_ eq 'get_user_by_code' } @provider_can ) {
+        like $res->content,
+        qr/Enter your username to obtain an email to reset your password/,
+        "... which has password reset option (reset_password_handler=>1).";
+    }
+    else {
+        unlike $res->content,
+        qr/Enter your username to obtain an email to reset your password/,
+        "... which has *no* password reset option (reset_password_handler=>0)";
+
+        # code coverage 'Reset password code submitted?' section of default
+        # get /login route.
+        $res = get('/login/12345678901234567890123456789012');
+        ok $res->is_success, "... and try a get /login/<code> is_success";
+    }
 
     # post empty /login
 
     $res = post('/login');
-    TODO: {
-        local $TODO = "POST /login returns 401 for failed login";
-        ok $res->is_success, "POST /login is_success";
-    }
+    ok $res->is_success, "POST /login is_success";
     like $res->content, qr/input.+name="password"/,
       "... and we have the login page";
     like $res->content, qr/LOGIN FAILED/,
@@ -1005,10 +1013,7 @@ sub _logged_in_user {
     # post empty /login
 
     $res = post('/login');
-    TODO: {
-        local $TODO = "POST /login returns 401 for failed login";
-        ok $res->is_success, "POST /login is_success";
-    }
+    ok $res->is_success, "POST /login is_success";
     like $res->content, qr/input.+name="password"/,
       "... and we have the login page";
     like $res->content, qr/LOGIN FAILED/,
@@ -1223,20 +1228,98 @@ sub _logged_in_user_password_expired {
 #------------------------------------------------------------------------------
 
 sub _password_reset {
-    my $res;
+    my ( $res, $code );
 
-    $res = get('/login');
-    ok 1;
-}
+    # request password reset with non-existant user
 
-#------------------------------------------------------------------------------
-#
-#  password_reset_send
-#
-#------------------------------------------------------------------------------
+    $Dancer2::Plugin::Auth::Extensible::Test::App::data = undef;
+
+    $res = post( '/login',
+        [ username_reset => 'NoSuchUser', submit_reset => 'truthy value' ] );
+
+    ok $res->is_success, "POST /login with password reset request is_success"
+      or diag explain $res;
+
+    like $res->content, qr/A password reset request has been sent/,
+      "... and we see \"A password reset request has been sent\" in page";
+
+    ok !defined $Dancer2::Plugin::Auth::Extensible::Test::App::data,
+      "... and password_reset_send_email was not called."
+      or diag explain $Dancer2::Plugin::Auth::Extensible::Test::App::data;
+
+    # call /login/$code with bad code
+
+    $res = get("/login/12345678901234567890123456789012");
+
+    ok $res->is_success, "GET /login/<code> with bad code is_success"
+      or diag explain $res;
+
+    like $res->content, qr/You need to log in to continue/,
+      "... and we have the /login page.";
+
+    # request password reset with valid user
+
+    $Dancer2::Plugin::Auth::Extensible::Test::App::data = undef;
+
+    $res = post( '/login',
+        [ username_reset => 'dave', submit_reset => 'truthy value' ] );
+
+    ok $res->is_success, "POST /login with password reset request is_success"
+      or diag explain $res;
+
+    like $res->content, qr/A password reset request has been sent/,
+      "... and we see \"A password reset request has been sent\" in page";
+
+    cmp_deeply $Dancer2::Plugin::Auth::Extensible::Test::App::data,
+      {
+        called => 1,
+        code   => re(qr/\w+/),
+        email  => ignore(),
+      },
+      "... and password_reset_send_email received code and email.";
+
+    $code = $Dancer2::Plugin::Auth::Extensible::Test::App::data->{code};
+
+    # get /login/$code
+
+    $trap->read;
+    $res = get("/login/$code");
+
+    ok $res->is_success, "GET /login/<code> with good code is_success"
+      or diag explain $res;
+
+    like $res->content,
+      qr/Please click the button below to reset your password/,
+      "... and we have the /login page with reset password link.";
+
+    # post /login/$code with bad code
+
+    $trap->read;
+    $res = post(
+        "/login/12345678901234567890123456789012",
+        [ confirm_reset => "Reset password" ]
+    );
+    ok $res->is_success, "POST /login/<code> with bad code is_success",
+      or diag explain $res;
+    unlike $res->content, qr/Your new password is \w{8}\</,
+      "... and we are NOT given a new password";
+    like $res->content, qr/LOGIN FAILED/, "... but see LOGIN FAILED.";
+
+    # post /login/$code with good code
+
+    $trap->read;
+    $res = post( "/login/$code", [ confirm_reset => "Reset password" ] );
+    ok $res->is_success, "POST /login/<code> with good code is_success",
+      or diag explain $res;
+    like $res->content, qr/Your new password is \w{8}\</,
+      "... and we are given a new password.";
+
+    # reset dave's password for later tests
+
+    $res =
+      post( '/user_password', [ username => 'dave', new_password => 'beer' ] );
+    is $res->content, "dave", "Reset dave's password to beer";
 
-sub _password_reset_send {
-    ok 1;
 }
 
 #------------------------------------------------------------------------------
@@ -1455,13 +1538,10 @@ sub _roles {
 
         my $res = get('/not_allroles');
 
-        is( $res->code, 302, "/not_allroles response code 302" )
+        is( $res->code, 403, "/not_allroles response code 403" )
           or diag explain $trap->read;
-        is(
-            $res->headers->header('Location'),
-            'http://localhost/login/denied?return_url=%2Fnot_allroles',
-            '/not_allroles redirected to denied page'
-        );
+        like $res->content, qr/Permission Denied/,
+          "... and we got the Permission Denied page.";
     }
 
     {
@@ -1481,15 +1561,11 @@ sub _roles {
 
         my $res = get('/piss');
 
-        is( $res->code, 302,
-            "Redirect on a route requiring a role we don't have" )
+        is( $res->code, 403,
+            "route requiring a role we don't have gets response code 403" )
           or diag explain $trap->read;
-
-        is(
-            $res->headers->header('Location'),
-            'http://localhost/login/denied?return_url=%2Fpiss',
-            "We cannot request a route requiring a role we don't have"
-        );
+        like $res->content, qr/Permission Denied/,
+          "... and we got the Permission Denied page.";
     }
 
     # 2 arg user_has_role
@@ -1839,79 +1915,196 @@ sub _update_user {
 #------------------------------------------------------------------------------
 
 sub _user_password {
-    {
-        $trap->read;    # clear logs
+    my $res;
 
-        my $res = get('/user_password?username=dave&password=beer');
-        is $res->code, 200,
-          "/user_password?username=dave&password=beer response is 200"
-          or diag explain $trap->read;
-        ok $res->content, "content shows success";
-    }
-    {
-        $trap->read;    # clear logs
+    # user_password with valid username and password, no realm
 
-        my $res = get('/user_password?username=dave&password=cider');
-        is $res->code, 200,
-          "/user_password?username=dave&password=cider response is 200"
-          or diag explain $trap->read;
-        ok !$res->content, "content shows fail";
-    }
-    {
-        $trap->read;    # clear logs
+    $trap->read;
+    $res = post( '/user_password', [ username => 'dave', password => 'beer' ] );
 
-        my $res =
-          get('/user_password?username=dave&password=beer&realm=config1');
+    ok $res->is_success,
+      "/user_password with valid username and password returns is_success"
+      or diag explain $trap->read;
+    is $res->content, 'dave', "... and it returned the username.";
 
-        is $res->code, 200,
-"/user_password?username=dave&password=beer&realm=config1 response is 200"
-          or diag explain $trap->read;
-        ok $res->content, "content shows success";
-    }
-    {
-        $trap->read;    # clear logs
+    # user_password with valid username but bad password, no realm
 
-        my $res =
-          get('/user_password?username=dave&password=beer&realm=config2');
+    $trap->read;
+    $res =
+      post( '/user_password', [ username => 'dave', password => 'BadPW' ] );
 
-        is $res->code, 200,
-"/user_password?username=dave&password=beer&realm=config2 response is 200"
-          or diag explain $trap->read;
-        ok !$res->content, "content shows fail";
-    }
+    ok $res->is_success,
+      "/user_password with valid username and password returns is_success"
+      or diag explain $trap->read;
+    ok !$res->content, "... and response is undef/empty.";
+
+    # user_password with valid username and password and realm
+
+    $trap->read;
+    $res = post( '/user_password',
+        [ username => 'dave', password => 'beer', realm => 'config1' ] );
+
+    ok $res->is_success,
+      "/user_password with valid username, password and realm is_success"
+      or diag explain $trap->read;
+    is $res->content, 'dave', "... and it returned the username.";
+
+    # user_password with valid username and password but wrong realm
+
+    $trap->read;
+    $trap->read;
+    $res = post( '/user_password',
+        [ username => 'dave', password => 'beer', realm => 'config2' ] );
+
+    ok $res->is_success,
+      "/user_password with valid username, password but bad realm is_success"
+      or diag explain $trap->read;
+    ok !$res->content, "content shows fail";
 
     # now with logged_in_user
 
-    {
-        $trap->read;    # clear logs
+    $trap->read;
+    $res = post( '/login', [ username => 'dave', password => 'beer' ] );
 
-        my $res = post( '/login', [ username => 'dave', password => 'beer' ] );
+    is( $res->code, 302, 'Login with real details succeeds' )
+      or diag explain $trap->read;
 
-        is( $res->code, 302, 'Login with real details succeeds' )
-          or diag explain $trap->read;
+    is get('/loggedin')->content, "You are logged in",
+      "... and checking /loggedin route shows we are logged in";
 
-        is get('/loggedin')->content, "You are logged in",
-          "... and checking /loggedin route shows we are logged in";
-    }
-    {
-        $trap->read;    # clear logs
+    # good password as only arg with logged in user
 
-        my $res = get('/user_password?password=beer');
-        is $res->code, 200, "/user_password?password=beer response is 200"
-          or diag explain $trap->read;
-        ok $res->content, "content shows success";
-    }
-    {
-        $trap->read;    # clear logs
+    $trap->read;
+    $res = post( '/user_password', [ password => 'beer' ] );
+    ok $res->is_success, "user_password password=beer is_success"
+      or diag explain $trap->read;
+    is $res->content, 'dave', "... and it returned the username.";
 
-        my $res = get('/user_password?password=cider');
-        is $res->code, 200, "/user_password?password=cider response is 200"
-          or diag explain $trap->read;
-        ok !$res->content, "content shows fail";
-    }
+    # bad password as only arg with logged in user
 
-    # cleanup
-    get('/logout');
+    $trap->read;
+    $res = post( '/user_password', [ password => 'cider' ] );
+    ok $res->is_success, "user_password password=cider is_success"
+      or diag explain $trap->read;
+    ok !$res->content, "content shows fail";
+
+    # logout
+
+    $res = get('/logout');
+    ok $res->is_redirect, "logout user dave is_redirect as expected";
+    ok get('/loggedin')->is_redirect, 
+      "... and checking /loggedin route shows dave is logged out.";
+
+    # search for user by code that no user has yet
+
+    my $code = 'UserPasswordResetCode123';
+    $trap->read;
+    $res = post( '/user_password', [ code => $code ] );
+    ok $res->is_success, "user_password with code no user has is_success"
+      or diag explain $trap->read;
+    ok !$res->content, "content shows fail";
+
+    # add code to dave's account details
+
+    $trap->read;
+    $res = post( '/update_user',
+        [ username => 'dave', realm => 'config1', pw_reset_code => $code ] );
+    ok $res->is_success,
+      "Add password reset code to dave's account details is_success.";
+
+    # now search for dave using code
+
+    $trap->read;
+    $res = post( '/user_password', [ code => $code ] );
+    ok $res->is_success, "user_password with code no user has is_success"
+      or diag explain $trap->read;
+    is $res->content, 'dave', "... and user dave was found.";
+
+    # change password
+
+    $trap->read;
+    $res = post( '/user_password',
+        [ username => 'dave', new_password => 'paleale' ] );
+    ok $res->is_success,
+      "Update password without giving old password is_success";
+    is $res->content, 'dave', "... and it returns the username."
+      or diag explain $trap->read;
+     
+    # try login with old password
+
+    $trap->read;
+    $res = post( '/login', [ username => 'dave', password => 'beer' ] );
+
+    ok $res->is_success, 'Login with old password fails with 200 OK code'
+      or diag explain $res;
+
+    ok get('/loggedin')->is_redirect,
+      "... and checking /loggedin route shows we are NOT logged in.";
+
+    # now new password
+
+    $trap->read;
+    $res = post( '/login', [ username => 'dave', password => 'paleale' ] );
+
+    is( $res->code, 302, 'Login with real details succeeds' )
+      or diag explain $trap->read;
+
+    is get('/loggedin')->content, "You are logged in",
+      "... and checking /loggedin route shows we are logged in";
+
+    # logout
+
+    $res = get('/logout');
+    ok $res->is_redirect, "logout user dave is_redirect as expected";
+    ok get('/loggedin')->is_redirect, 
+      "... and checking /loggedin route shows dave is logged out.";
+
+
+    # try to change password but supply bad old password
+
+    $trap->read;
+    $res = post( '/user_password',
+        [ username => 'dave', password => 'bad', new_password => 'beer' ] );
+    ok $res->is_success, "Update password with bad old password is_success";
+    ok !$res->content, "... and it returns false."
+      or diag explain $trap->read;
+     
+    # try to change password and supply good old password
+
+    $trap->read;
+    $res = post( '/user_password',
+        [ username => 'dave', password => 'paleale', new_password => 'beer' ] );
+    ok $res->is_success, "Update password with good old password is_success";
+    is $res->content, 'dave', "... and user dave was found.";
+     
+    # try login with old password
+
+    $trap->read;
+    $res = post( '/login', [ username => 'dave', password => 'paleale' ] );
+
+    ok $res->is_success, 'Login with old password fails with 200 OK code'
+      or diag explain $res;
+
+    ok get('/loggedin')->is_redirect,
+      "... and checking /loggedin route shows we are NOT logged in.";
+
+    # now new password
+
+    $trap->read;
+    $res = post( '/login', [ username => 'dave', password => 'beer' ] );
+
+    is( $res->code, 302, 'Login with real details succeeds' )
+      or diag explain $trap->read;
+
+    is get('/loggedin')->content, "You are logged in",
+      "... and checking /loggedin route shows we are logged in";
+
+    # logout
+
+    $res = get('/logout');
+    ok $res->is_redirect, "logout user dave is_redirect as expected";
+    ok get('/loggedin')->is_redirect, 
+      "... and checking /loggedin route shows dave is logged out.";
 }
 
 1;
