@@ -6,7 +6,7 @@ Dancer2::Plugin::Auth::Extensible::Test::App - Dancer2 app for testing providers
 
 =cut
 
-our $VERSION = '0.622';
+our $VERSION = '0.700';
 
 use strict;
 use warnings;
@@ -26,13 +26,14 @@ set show_errors => 1;
 # nasty shared global makes it easy to pass data between app and test script
 our $data = {};
 
-config->{plugins}->{"Auth::Extensible"}->{reset_password_handler} = 1;
+config->{plugins}->{"Auth::Extensible"}->{password_reset_send_email} =
+  __PACKAGE__ . "::email_send";
 config->{plugins}->{"Auth::Extensible"}->{welcome_send} =
-  __PACKAGE__ . "::welcome_send";
+  __PACKAGE__ . "::email_send";
 
-sub welcome_send {
+sub email_send {
     my ( $plugin, %args ) = @_;
-    $data = \%args;
+    $data = { %args, called => 1 };
 }
 
 # we need the plugin object and a provider for provider tests
@@ -43,6 +44,9 @@ my @provider_can = ();
 
 push @provider_can, 'record_lastlogin' if $plugin->config->{record_lastlogin};
 
+config->{plugins}->{"Auth::Extensible"}->{reset_password_handler} = 1
+  if $provider->can('get_user_by_code');
+
 #
 # IMPORTANT NOTE
 #
@@ -618,6 +622,9 @@ get '/does_undef_drink_beer' => sub {
 get '/user_password' => sub {
     return user_password params('query');
 };
+post '/user_password' => sub {
+    return user_password %{ body_parameters->as_hashref };
+};
 
 get '/update_current_user' => sub {
     my $user = update_current_user name => "I love cider";
